package sun.awt.geom;

import java.util.Arrays;
import java.util.Enumeration;
import java.util.Vector;
import java.util.Comparator;

public abstract class AreaOp
{
    public static final int CTAG_LEFT = 0;
    public static final int CTAG_RIGHT = 1;
    public static final int ETAG_IGNORE = 0;
    public static final int ETAG_ENTER = 1;
    public static final int ETAG_EXIT = -1;
    public static final int RSTAG_INSIDE = 1;
    public static final int RSTAG_OUTSIDE = -1;
    private static Comparator YXTopComparator;
    private static CurveLink[] EmptyLinkList;
    private static ChainEnd[] EmptyChainList;
    
    private AreaOp() {
    }
    
    public abstract void newRow();
    
    public abstract int classify(Edge p0);
    
    public abstract int getState();
    
    public Vector calculate(Vector vector, Vector vector2) {
        Vector<Object> vector3 = new Vector();
        addEdges(vector3, vector, 0);
        addEdges(vector3, vector2, 1);
        return this.pruneEdges(vector3);
    }
    
    private static void addEdges(Vector vector, Vector vector2, int n) {
        Enumeration<Object> elements = vector2.elements();
        while (elements.hasMoreElements()) {
            Curve curve = (Curve)elements.nextElement();
            if (curve.getOrder() > 0) {
                vector.add(new Edge(curve, n));
            }
            else {
                continue;
            }
        }
    }
    
    private Vector pruneEdges(Vector vector) {
        int size = vector.size();
        if (size < 2) {
            return vector;
        }
        Edge[] a = (Edge[])(Edge[])vector.toArray(new Edge[size]);
        Arrays.sort(a, AreaOp.YXTopComparator);
        int i = 0;
        int n = 0;
        double[] array = new double[2];
        Vector<Object> vector2 = new Vector();
        Vector<Object> vector3 = new Vector();
        Vector<CurveLink> vector4 = new Vector();
        while (i < size) {
            double yTop = array[0];
            int j;
            int n2;
            for (n2 = (j = n - 1); j >= i; --j) {
                Edge edge = a[j];
                if (edge.getCurve().getYBot() > yTop) {
                    if (n2 > j) {
                        a[n2] = edge;
                    }
                    --n2;
                }
            }
            i = n2 + 1;
            if (i >= n) {
                if (n >= size) {
                    break;
                }
                yTop = a[n].getCurve().getYTop();
                if (yTop > array[0]) {
                    finalizeSubCurves(vector2, vector3);
                }
                array[0] = yTop;
            }
            while (n < size && a[n].getCurve().getYTop() <= yTop) {
                ++n;
            }
            array[1] = a[i].getCurve().getYBot();
            if (n < size) {
                double yTop2 = a[n].getCurve().getYTop();
                if (array[1] > yTop2) {
                    array[1] = yTop2;
                }
            }
            int n3 = 1;
            Edge edge2;
            for (; i < n; a[i] = edge2, ++i) {
                edge2 = a[i];
                edge2.setEquivalence(0);
                while (i > i) {
                    Edge edge3 = a[i - 1];
                    int compareTo = edge2.compareTo(edge3, array);
                    if (array[1] <= array[0]) {
                        throw new InternalError("backstepping to " + array[1] + " from " + array[0]);
                    }
                    if (compareTo >= 0) {
                        if (compareTo == 0) {
                            int equivalence = edge3.getEquivalence();
                            if (equivalence == 0) {
                                equivalence = n3++;
                                edge3.setEquivalence(equivalence);
                            }
                            edge2.setEquivalence(equivalence);
                            break;
                        }
                        else {
                            break;
                        }
                    }
                    else {
                        a[i] = edge3;
                        --i;
                    }
                }
            }
            this.newRow();
            double n4 = array[0];
            double n5 = array[1];
            for (; i < n; ++i) {
                Edge edge4 = a[i];
                int equivalence2 = edge4.getEquivalence();
                int classify;
                if (equivalence2 != 0) {
                    int state = this.getState();
                    classify = ((state == 1) ? -1 : 1);
                    Edge edge5 = null;
                    do {
                        this.classify(edge4);
                        if (edge5 == null && edge4.isActiveFor(n4, classify)) {
                            edge5 = edge4;
                        }
                        double yBot = edge4.getCurve().getYBot();
                        if (yBot > n5) {
                            n5 = yBot;
                        }
                        else {
                            continue;
                        }
                    } while (++i < n && (edge4 = a[i]).getEquivalence() == equivalence2);
                    --i;
                    if (this.getState() == state) {
                        classify = 0;
                    }
                    else {
                        edge4 = ((edge5 != null) ? edge5 : edge4);
                    }
                }
                else {
                    classify = this.classify(edge4);
                }
                if (classify != 0) {
                    edge4.record(n5, classify);
                    vector4.add(new CurveLink(edge4.getCurve(), n4, n5, classify));
                }
            }
            if (this.getState() != -1) {
                System.out.println("Still inside at end of active edge list!");
                System.out.println("num curves = " + (n - i));
                System.out.println("num links = " + vector4.size());
                System.out.println("y top = " + array[0]);
                if (n < size) {
                    System.out.println("y top of next curve = " + a[n].getCurve().getYTop());
                }
                else {
                    System.out.println("no more curves");
                }
                for (; i < n; ++i) {
                    Edge x = a[i];
                    System.out.println(x);
                    int equivalence3 = x.getEquivalence();
                    if (equivalence3 != 0) {
                        System.out.println("  was equal to " + equivalence3 + "...");
                    }
                }
            }
            resolveLinks(vector2, vector3, vector4);
            vector4.clear();
            array[0] = n5;
        }
        finalizeSubCurves(vector2, vector3);
        Vector o = new Vector();
        Enumeration<Object> elements = vector2.elements();
        while (elements.hasMoreElements()) {
            CurveLink curveLink = (CurveLink)elements.nextElement();
            o.add(curveLink.getMoveto());
            CurveLink next = curveLink;
            while ((next = next.getNext()) != null) {
                if (curveLink.absorb(next)) {
                    continue;
                }
                o.add(curveLink.getSubCurve());
                curveLink = next;
            }
            o.add(curveLink.getSubCurve());
        }
        return o;
    }
    
    public static void finalizeSubCurves(Vector vector, Vector vector2) {
        int size = vector2.size();
        if (size == 0) {
            return;
        }
        if ((size & 0x1) != 0x0) {
            throw new InternalError("Odd number of chains!");
        }
        ChainEnd[] a = new ChainEnd[size];
        vector2.toArray(a);
        for (int i = 1; i < size; i += 2) {
            CurveLink linkTo = a[i - 1].linkTo(a[i]);
            if (linkTo != null) {
                vector.add(linkTo);
            }
        }
        vector2.clear();
        return;
    }
    
    public static void resolveLinks(Vector vector, Vector vector2, Vector vector3) {
        int size = vector3.size();
        CurveLink[] emptyLinkList;
        if (size == 0) {
            emptyLinkList = AreaOp.EmptyLinkList;
        }
        else {
            if ((size & 0x1) != 0x0) {
                throw new InternalError("Odd number of new curves!");
            }
            emptyLinkList = new CurveLink[size + 2];
            vector3.toArray(emptyLinkList);
        }
        int size2 = vector2.size();
        ChainEnd[] emptyChainList;
        if (size2 == 0) {
            emptyChainList = AreaOp.EmptyChainList;
        }
        else {
            if ((size2 & 0x1) != 0x0) {
                throw new InternalError("Odd number of chains!");
            }
            emptyChainList = new ChainEnd[size2 + 2];
            vector2.toArray(emptyChainList);
        }
        int n = 0;
        int n2 = 0;
        vector2.clear();
        ChainEnd e = emptyChainList[0];
        ChainEnd chainEnd = emptyChainList[1];
        CurveLink curveLink = emptyLinkList[0];
        CurveLink curveLink2 = emptyLinkList[1];
        while (e != null || curveLink != null) {
            int n3 = curveLink == null;
            int n4 = e == null;
            if (n3 == 0 && n4 == 0) {
                n3 = ((n & 0x1) == 0x0 && e.getX() == chainEnd.getX());
                n4 = ((n2 & 0x1) == 0x0 && curveLink.getX() == curveLink2.getX());
                if (n3 == 0 && n4 == 0) {
                    double x = e.getX();
                    double x2 = curveLink.getX();
                    n3 = (chainEnd != null && x < x2 && obstructs(chainEnd.getX(), x2, n));
                    n4 = (curveLink2 != null && x2 < x && obstructs(curveLink2.getX(), x, n2));
                }
            }
            if (n3 != 0) {
                CurveLink linkTo = e.linkTo(chainEnd);
                if (linkTo != null) {
                    vector.add(linkTo);
                }
                n += 2;
                e = emptyChainList[n];
                chainEnd = emptyChainList[n + 1];
            }
            if (n4 != 0) {
                ChainEnd e2 = new ChainEnd(curveLink, null);
                ChainEnd o = new ChainEnd(curveLink2, e2);
                e2.setOtherEnd(o);
                vector2.add(e2);
                vector2.add(o);
                n2 += 2;
                curveLink = emptyLinkList[n2];
                curveLink2 = emptyLinkList[n2 + 1];
            }
            if (n3 == 0 && n4 == 0) {
                e.addLink(curveLink);
                vector2.add(e);
                ++n;
                e = chainEnd;
                chainEnd = emptyChainList[n + 1];
                ++n2;
                curveLink = curveLink2;
                curveLink2 = emptyLinkList[n2 + 1];
            }
            else {
                continue;
            }
        }
        if ((vector2.size() & 0x1) != 0x0) {
            System.out.println("Odd number of chains!");
        }
    }
    
    public static boolean obstructs(double n, double n2, int n3) {
        return ((n3 & 0x1) == 0x0) ? (n <= n2) : (n < n2);
    }
    
    AreaOp(AreaOp.1 comparator) {
        this();
    }
    
    static {
        AreaOp.YXTopComparator = new Comparator() {
            @Override
            public int compare(Object o, Object o2) {
                Curve curve = ((Edge)o).getCurve();
                Curve curve2 = ((Edge)o2).getCurve();
                double n = curve.getYTop();
                double n2;
                if (n == (n2 = curve2.getYTop()) && (n = curve.getXTop()) == (n2 = curve2.getXTop())) {
                    return 0;
                }
                if (n < n2) {
                    return -1;
                }
                return 1;
            }
        };
        AreaOp.EmptyLinkList = new CurveLink[2];
        AreaOp.EmptyChainList = new ChainEnd[2];
    }
    
    public abstract static class CAGOp extends AreaOp
    {
        boolean inLeft;
        boolean inRight;
        boolean inResult;
        
        public CAGOp() {
            super(null);
        }
        
        @Override
        public void newRow() {
            this.inLeft = false;
            this.inRight = false;
            this.inResult = false;
        }
        
        @Override
        public int classify(Edge edge) {
            if (edge.getCurveTag() == 0) {
                this.inLeft = !this.inLeft;
            }
            else {
                this.inRight = !this.inRight;
            }
            int classification = this.newClassification(this.inLeft, this.inRight);
            if ((this.inResult ? 1 : 0) == classification) {
                return 0;
            }
            this.inResult = classification;
            return classification ? 1 : -1;
        }
        
        @Override
        public int getState() {
            return this.inResult ? 1 : -1;
        }
        
        public abstract boolean newClassification(boolean p0, boolean p1);
    }
    
    public static class AddOp extends AreaOp.CAGOp
    {
        @Override
        public boolean newClassification(boolean b, boolean b2) {
            return b || b2;
        }
    }
    
    public static class SubOp extends AreaOp.CAGOp
    {
        @Override
        public boolean newClassification(boolean b, boolean b2) {
            return b && !b2;
        }
    }
    
    public static class IntOp extends AreaOp.CAGOp
    {
        @Override
        public boolean newClassification(boolean b, boolean b2) {
            return b && b2;
        }
    }
    
    public static class XorOp extends AreaOp.CAGOp
    {
        @Override
        public boolean newClassification(boolean b, boolean b2) {
            return b != b2;
        }
    }
    
    public static class NZWindOp extends AreaOp
    {
        private int count;
        
        public NZWindOp() {
            super(null);
        }
        
        @Override
        public void newRow() {
            this.count = 0;
        }
        
        @Override
        public int classify(Edge edge) {
            int count = this.count;
            int n = count == 0;
            int count2 = count + edge.getCurve().getDirection();
            this.count = count2;
            return (count2 == 0) ? -1 : n;
        }
        
        @Override
        public int getState() {
            return (this.count == 0) ? -1 : 1;
        }
    }
    
    public static class EOWindOp extends AreaOp
    {
        private boolean inside;
        
        public EOWindOp() {
            super(null);
        }
        
        @Override
        public void newRow() {
            this.inside = false;
        }
        
        @Override
        public int classify(Edge edge) {
            int inside = !this.inside;
            this.inside = inside;
            return inside ? 1 : -1;
        }
        
        @Override
        public int getState() {
            return this.inside ? 1 : -1;
        }
    }
}
